@use "ct-card.scss";
@import "@patternfly/patternfly/base/patternfly-variables.scss";

/* Style the list cards as ct-cards */
.pf-v6-c-page__main-section .pf-v6-c-card {
    @extend .ct-card;
}

// Expand the link to the container, for easier clickability
#virtual-machines-listing th a {
    display: block;
    text-align: start;
}

// Add spacing between buttons in the toolbar
.pf-v6-c-toolbar__item {
    gap: 0 var(--pf-t--global--spacer--xs);
}
