.footer-warning {
    color: var(--pf-t--global--text--color--status--danger--default);
    margin-block-end: var(--pf-t--global--spacer--sm);
}

.popover-headline {
    margin-block-end: var(--pf-t--global--spacer--xs);
}

/* Add some padding to avoid scrollbar being shown */
#migrate-modal form {
    padding-block-end: var(--pf-t--global--spacer--xs);
}
