@use "_global-variables.scss" as *;

/* Rely on the margin from the Card for spacing */
.ct-card.pf-v6-c-card .table {
  margin-block-end: 0;
}

.ct-card.pf-v6-c-card .pf-v6-c-card__title-text {
  font-weight: normal;
  font-size: var(--pf-t--global--font--size--2xl);
}

.ct-cards-grid {
  --ct-grid-columns: 2;
  --pf-v6-l-gallery--GridTemplateColumns: repeat(var(--ct-grid-columns), 1fr);

  > .pf-v6-c-card:not(.ct-card-info) {
    // Extend all non-info cards to be full width;
    // let ct-card-info fit 1 column of the grid
    grid-column: 1 / -1;
  }

  @media screen and (max-width: $pf-v6-global--breakpoint--lg) {
    // Shrink to 1 column when space is constrained
    --ct-grid-columns: 1;
  }
}

// Remove redundant padding from embedded toolbars (handled by header)
// Toolbars in card headers are not a common scenario so no need to upstream this
.ct-card.pf-v6-c-card .pf-v6-c-toolbar,
.ct-card.pf-v6-c-card .pf-v6-c-toolbar__content {
  padding: 0;
}
