@import "ct-card.scss";

.vm-details .pf-v6-c-card {
    @extend .ct-card;
}

.consoles-card h2 button {
    font-size: var(--pf-t--global--font--size--2xl);
    padding: 0;

    > .pf-v6-c-button__icon {
        font-size: var(--pf-v6-c-button--FontSize);
    }
}

.vm-top-panel {
  display: flex;
  flex-wrap: wrap;

  .vm-name {
    margin-inline-end: 1.5rem;
  }

  &,
  .btn-group {
    align-items: center;
  }
}

.pf-v6-l-gallery.ct-vm-overview {
  $ctVmBreakpoint: 1000px;

  @media screen and (max-width: $ctVmBreakpoint) {
    --pf-v6-l-gallery--GridTemplateColumns: 1fr;
  }

  @media screen and (min-width: $ctVmBreakpoint + 1) {
    --pf-v6-l-gallery--GridTemplateColumns: minmax(max-content, 20rem) 1fr;

    .consoles-card {
        grid-row: 1 / span 2;
        grid-column: 2 / -1;
    }
  }

  .networks-card, .disks-card, .snapshots-card, .hostdevs-card, .filesystems-card {
    grid-column: 1 / -1;
  }

  .pf-v6-c-card {
    &__header {
      font-size: var(--pf-t--global--font--size--xl);
    }

    &__footer {
      &:empty {
        display: none;
      }
    }
  }

  .pf-v6-c-progress {
    --pf-v6-c-progress--GridGap: var(--pf-t--global--spacer--xs);
  }
}
