// Add some space between resource status and possible error state
.resource-state-text .pf-v6-c-label__text {
    display: flex;
    gap: var(--pf-t--global--spacer--xs);
}

.resource-state-text:first-child {
  // In definition lists (for Overview),
  // match text size and alignment
  .pf-v6-c-table &,
  .pf-v6-c-description-list & {
    // Match text size
    font-size: inherit;
    // Compensate for padding
    position: relative;
    // Using margin to remove the extra padded space,
    // instead of keeping the space and moving contents w/ just left/top
    margin-inline-start: calc(-1 * var(--pf-v6-c-label--PaddingInlineStart));
  }
}

.pf-v6-c-label.resource-state--shut-off, .pf-v6-c-label.resource-state--inactive {
  background: transparent;
}
