@use "global-variables" as *;
@use "@patternfly/patternfly/utilities/Text/text.scss";

.ct-password-strength-meter {
  grid-gap: var(--pf-t--global--spacer--xs);
  inline-size: var(--pf-t--global--spacer--2xl);

  .pf-v6-c-progress__description, .pf-v6-c-progress__status {
    display: none;
  }
}
