@forward "@patternfly/patternfly/sass-utilities";

/*
 * PatternFly 4 adapting the lists too early.
 * When PF4 has a breakpoint of 768px width, it's actually 1108 for us, as the sidebar is 340px.
 * (This does use the intended content area, but there's a mismatch between content and browser width as we use iframes.)
 * So redefine grid breakpoints
 */
// REVIEW @Venefilyn: Check if still relevant, maybe use rem instead
$pf-v6-global--breakpoint--xs: 0 !default;
// Do not override the sm breakpoint as for width < 768px the left nav is hidden
$pf-v6-global--breakpoint--md: 428px !default;
$pf-v6-global--breakpoint--lg: 652px !default;
$pf-v6-global--breakpoint--xl: 860px !default;
$pf-v6-global--breakpoint--2xl: 1110px !default;
